{$S-,R-,V-,I-,B-}

unit tpEdit;
  {-Line editor and keyboard input routines}

interface

uses
  TpCrt,
  TpCmd,
  TpString;

  {.F-}
const
  RSnone      = 00; {Not a command}
  RSchar      = 01; {A character to enter the string}
  RSctrlChar  = 02; {Accept control character}
  RSenter     = 03; {Accept current string and quit}
  RSquit      = 04; {Restore default string and quit}
  RSrestore   = 05; {Restore default and continue}
  RShome      = 06; {Cursor to beginning of line}
  RSend       = 07; {Cursor to end of line}
  RSleft      = 08; {Cursor left by one character}
  RSright     = 09; {Cursor right by one character}
  RSwordLeft  = 10; {Cursor left one word}
  RSwordRight = 11; {Cursor right one word}
  RSback      = 12; {Backspace one character}
  RSdel       = 13; {Delete current character}
  RSdelEol    = 14; {Delete from cursor to end of line}
  RSdelBol    = 15; {Delete from beginning of line to cursor}
  RSdelLine   = 16; {Delete entire line}
  RSdelWord   = 17; {Delete word to right of cursor}
  RSins       = 18; {Toggle insert mode}
  RShelp      = 19; {Invoke help routine}
  RSuser0     = 20; {user-defined exit commands}
  RSuser1     = 21;
  RSuser2     = 22;
  RSuser3     = 23;
  RSuser4     = 24;
  RSuser5     = 25;
  RSuser6     = 26;
  RSuser7     = 27;
  RSuser8     = 28;
  RSuser9     = 29;
{.F+}

type
  RStype = RSnone..RSuser9;
  CharSet = set of Char;

const
  {the following govern the behavior of ReadString and the numeric reads}
  ForceUpper = False; {force chars to uppercase?}
  CursorToEnd = True; {start cursor at end of string? else at start}
  TrimBlanks  = True; {trim leading/trailing blanks? -- forced on
                                temporarily for the numeric reads}
  InsertByDefault  = True; {default to insert mode?}
  EditSize : Byte = 0;       {if NOT zero, this overrides MaxLen as edit field width}
  ClearFirstChar = True; {clear string if first char entered is ASCII}
  PasswordMode = False; {suppress display of actual string?}
  PasswordChar = ^G;  {character used to represent text in password mode}
  EditMapCtrls = True; {map control characters to alphabetics?}
  RSpadChar = ' ';    {character used to pad strings being edited}
  HouseCursorAtEnd = True; {extra character of width to hold cursor
                                      when string is full?}
  init_col:byte=0;



var
  EditKeyPtr : Pointer;      {pointer to routine to return next keystroke}
  RSCommand : RStype;        {last command given in ReadString}
  RSChWord : Word;           {last key entered in ReadString}
  cur_str:string;


procedure ReadString(Prompt : string;
                     Row, Col, MaxLen : Byte;
                     PromptAttr, StringAttr, CtrlAttr : Byte;
                     var Escaped : Boolean;
                     var S : string);
  {-Prompt for and receive a string}


function AddEditCommand(Cmd : RStype; NumKeys : Byte; Key1, Key2 : Word) : Boolean;
  {-Add a new command key assignment or change an existing one}

const
  {Keystroke to command mapping}
  EditKeyMax = 150;   {last available slot in EditKeySet}
  {ID string for installation programs}
  EditKeyID : string[16] = 'tpedit key array';
  {default key assignments}
  EditKeySet : array[0..EditKeyMax] of Byte = (
   {length keys         command type      key sequence}
    3,     $00, $00,    RSquit,          {^Break}
    3,     $00, $3B,    RShelp,          {F1}
    3,     $00, $47,    RShome,          {Home}
    3,     $00, $4B,    RSleft,          {Left}
    3,     $00, $4D,    RSright,         {Right}
    3,     $00, $4F,    RSend,           {End}
    3,     $00, $52,    RSins,           {Ins}
    3,     $00, $53,    RSdel,           {Del}
    3,     $00, $73,    RSwordLeft,      {^Left}
    3,     $00, $74,    RSwordRight,     {^Right}
    3,     $00, $75,    RSdelEol,        {^End}
    3,     $00, $77,    RSdelBol,        {^Home}
    2,     $01,         RSwordLeft,      {^A}
    2,     $04,         RSright,         {^D}
    2,     $06,         RSwordRight,     {^F}
    2,     $07,         RSdel,           {^G}
    2,     $08,         RSback,          {^H, Bksp}
    2,     $0D,         RSenter,         {^M, Enter}
    2,     $10,         RSctrlChar,      {^P}
    2,     $12,         RSrestore,       {^R}
    2,     $13,         RSleft,          {^S}
    2,     $14,         RSdelWord,       {^T}
    2,     $18,         RSdelLine,       {^X}
    2,     $19,         RSdelLine,       {^Y}
    2,     $1B,         RSquit,          {Esc}
    2,     $7F,         RSback,          {^Bksp}
    3,     $11, $04,    RSend,           {^Q^D}
    3,     $11, $0C,    RSrestore,       {^Q^L}
    3,     $11, $13,    RShome,          {^Q^S}
    3,     $11, $19,    RSdelEol,        {^Q^Y}
                   0, 0, 0, 0, 0,        {110}
    0, 0, 0, 0, 0, 0, 0,
    {-----------pad to end of array----------}
                         0, 0, 0,        {120}
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {130}
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {140}
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0);       {150}
{.F+}

  {==========================================================================}

implementation

type
  String80 = string[80];
const
  NumError : string[32] = 'Invalid number. Press any key...';


  function GetKey : Word;
    {-Call routine pointed to by EditKeyPtr}
  inline(
    $FF/$1E/>EditKeyPtr);    {call dword ptr [>EditKeyPtr]}


  procedure ReadString(Prompt : string;
                       Row, Col, MaxLen : Byte;
                       PromptAttr, StringAttr, CtrlAttr : Byte;
                       var Escaped : Boolean;
                       var S : string);
    {-Prompt for and receive a string}
  var
    Ch : Char absolute RSChWord;
    St : string absolute cur_str;
    StLen : Byte absolute St;
    CursorSL : Word;
    CursorXY : Word;
    StCol : Byte;
    Sp : Byte;
    DelEnd : Byte;
    Inserting : Boolean;
    FirstChar : Boolean;
    SaveBreak : Boolean;
    Done : Boolean;
    StOffset : Byte;
    EditLen : Byte;
    MaxBoxLen : Byte;
    RSC : RStype absolute RSCommand;
    {$IFDEF UseMouse}
    SaveWaitState : Boolean;
    SaveMouse : Boolean;
    {$ENDIF}


    procedure ToggleInsertMode;
     {-Toggle between insert and overtype mode, keeping BIOS keyboard flag up
       to date}
    var
      BiosKbdFlag : Byte absolute $0040 : $0017;
    begin
      {toggle insert flag}
      Inserting := not Inserting;

      {use fat cursor if inserting}
      if Inserting then begin
        FatCursor;
        BiosKbdFlag := BiosKbdFlag or $80;
      end
      else begin
        NormalCursor;
        BiosKbdFlag := BiosKbdFlag and $7F;
      end;
    end;

    procedure DrawString;
      {-Draw the string}
    var
      A, I, SaveStLen : Byte;
      C : Char;
    begin
      {pad the end of the string}
      FillChar(St[Succ(StLen)], MaxBoxLen-StLen, RSpadChar);

      {make sure padding is drawn}
      SaveStLen := StLen;
      StLen := EditLen;


      {draw the string and restore the length byte}
      for I := 1 to StLen do begin
          C := St[StOffset+I];
          if (C < ' ') then begin
            if EditMapCtrls then
              C := Chr(Ord(C) or $40);
            A := CtrlAttr;
          end
          else
            A := StringAttr;
          FastWrite(C, Row, StCol+Pred(I), A);
      end;
      StLen := SaveStLen;

    end;

  begin
    {Determine if EditSize is appropriate here}
    if MaxLen > 254 then
      MaxLen := 254;
    if HouseCursorAtEnd then
      MaxBoxLen := MaxLen+1
    else
      MaxBoxLen := MaxLen;
    if (EditSize = 0) or (EditSize >= MaxBoxLen) then
      EditLen := MaxBoxLen
    else
      EditLen := EditSize;

    {Store cursor position and shape}
    GetCursorState(CursorXY, CursorSL);

    {Save break checking state}
    SaveBreak := CheckBreak;
    CheckBreak := False;
    Escaped := False;
    StOffset := 0;


    {Write prompt}
      FastWrite(Prompt, Row, Col, PromptAttr);
    StCol := Col+Length(Prompt);


    {Get the default string}
    St := S;
    if StLen > MaxLen then
      StLen := MaxLen;
    if init_col>0 then begin
     sp:=init_col;
     init_col:=0
    end else
    if CursorToEnd then
      Sp := Succ(StLen)
    else
      Sp := 1;
    FirstChar := True;

    {default to insert mode on if InsertByDefault is true}
    Inserting := not InsertByDefault;
    ToggleInsertMode;

    {Loop reading keys}
    Done := False;
    repeat
      {Position cursor and redraw string}
      if Sp > MaxLen then
        if HouseCursorAtEnd then
          Sp := Succ(MaxLen)
        else
          Sp := MaxLen;
      if Sp > EditLen+StOffset then
        StOffset := Sp-EditLen
      else if Sp < Succ(StOffset) then
        StOffset := Pred(Sp);
      DrawString;
        GoToXYAbs(StCol+Pred(Sp)-StOffset, Row);


      RSC := GetCommand(EditKeySet, EditKeyPtr, RSChWord);


      {deal with control characters if desired}
      if RSC = RSctrlChar then
        {don't allow control characters if attributes are the same}
        if (CtrlAttr = StringAttr) and EditMapCtrls then
          RSC := RSnone
        else begin
          BlockCursor;
          RSChWord := GetKey;
          RSC := RSchar;
          if Inserting then
            FatCursor
          else
            NormalCursor;
        end;

      if (RSC = RSchar) and ForceUpper then
        Ch := Upcase(Ch);

      {allow editing of the existing string}
      if FirstChar then begin
        FirstChar := False;

        if ClearFirstChar then
          {if first key is a character, clear the input string}
          if (RSC = RSchar) then begin
            StOffset := 0;
            StLen := 0;
            Sp := 1;
          end;
      end;

      case RSC of
        RSchar :             {A character to enter the string}
          if not Inserting then begin
            {overtype mode}
            if Sp <= MaxLen then begin
              St[Sp] := Ch;
              if Sp > StLen then
                StLen := Sp;
              Inc(Sp);
            end;
          end
          else
            {insert mode}
            if StLen < MaxLen then begin
              Insert(Ch, St, Sp);
              Inc(Sp);
            end;

        RSuser0..RSuser9,    {user-defined exit commands}
        RSenter :            {Accept current string and quit}
          Done := True;

        RSquit :             {Restore default string and quit}
          begin
            St := S;
            if CursorToEnd then
              Sp := Succ(StLen)
            else
              Sp := 1;
            if StLen > MaxLen then
              StLen := MaxLen;
            Done := True;
            Escaped := True;
          end;

        RShome :             {Cursor to beginning of line}
          Sp := 1;

        RSend :              {Cursor to end of line}
          Sp := Succ(StLen);

        RSdelEol :           {Delete from cursor to end of line}
          St := Copy(St, 1, Pred(Sp));

        RSdelBol :           {Delete from beginning of line to the cursor}
          begin
            Delete(St, 1, Pred(Sp));
            Sp := 1;
          end;

        RSdelLine :          {Delete entire line}
          begin
            StLen := 0;
            Sp := 1;
          end;

        RSrestore :          {Restore default and continue}
          begin
            St := S;
            if StLen > MaxLen then
              StLen := MaxLen;
            if CursorToEnd then
              Sp := Succ(StLen)
            else
              Sp := 1;
          end;

        RSleft :             {Cursor left by one character}
          if Sp > 1 then
            Dec(Sp);

        RSright :            {Cursor right by one character}
          if Sp <= StLen then
            Inc(Sp);

        RSwordLeft :         {Cursor left one word}
          if Sp > 1 then begin
            Dec(Sp);
            while (Sp >= 1) and ((Sp > StLen) or (St[Sp] = ' ')) do
              Dec(Sp);
            while (Sp >= 1) and (St[Sp] <> ' ') do
              Dec(Sp);
            Inc(Sp);
          end;

        RSwordRight :        {Cursor right one word}
          if Sp <= StLen then begin
            Inc(Sp);
            while (Sp <= StLen) and (St[Sp] <> ' ') do
              Inc(Sp);
            while (Sp <= StLen) and (St[Sp] = ' ') do
              Inc(Sp);
          end;

        RSdel :              {Delete current character}
          if Sp <= StLen then
            Delete(St, Sp, 1);

        RSback :             {Backspace one character}
          if Sp > 1 then begin
            Dec(Sp);
            Delete(St, Sp, 1);
            if StOffset > 0 then
              {String horizontally scrolled}
              if StOffset+EditLen >= StLen then
                {The rightmost portion of the string is displayed, so scroll}
                Dec(StOffset);
          end;

        RSdelWord :          {Delete word to right of cursor}
          if Sp <= StLen then begin
            {start deleting at the cursor}
            DelEnd := Sp;

            {delete all of the current word, if any}
            if St[Sp] <> ' ' then
              while (St[DelEnd] <> ' ') and (DelEnd <= StLen) do
                Inc(DelEnd);

            {delete any spaces prior to the next word, if any}
            while (St[DelEnd] = ' ') and (DelEnd <= StLen) do
              Inc(DelEnd);

            Delete(St, Sp, DelEnd-Sp);
          end;

        RSins :              {Toggle insert mode}
          ToggleInsertMode;

      end;
    until Done;

    {draw the string one last time}
    if Sp > MaxLen then
      if HouseCursorAtEnd then
        Sp := Succ(MaxLen)
      else
        Sp := MaxLen;
    if Sp > EditLen+StOffset then
      StOffset := Sp-EditLen
    else if Sp < Succ(StOffset) then
      StOffset := Pred(Sp);
    DrawString;

      S := St;

    {restore break checking status}
    CheckBreak := SaveBreak;

    {Restore cursor position and shape}
    RestoreCursorState(CursorXY, CursorSL);

  end;


  function AddEditCommand(Cmd : RStype; NumKeys : Byte; Key1, Key2 : Word) : Boolean;
    {-Add a new command key assignment or change an existing one}
  begin
    AddEditCommand :=
    AddCommandPrim(EditKeySet, EditKeyMax, Cmd, NumKeys, Key1, Key2);
  end;


begin
  {initialize procedure pointers}
  EditKeyPtr := @ReadKeyWord;
end.
